import torch.nn as nn
import torch
from utils import softmax_helper
from utils import init_weights
import numpy as np


class ConvDropoutNormLReLU(nn.Module):
    def __init__(self, input_channels, output_channels, dilation=1, kernel_size=3, padding=1, dropout=0.0, stride=1, conv_op=nn.Conv2d,
                 norm_op=nn.BatchNorm2d, dropout_op=nn.Dropout2d, inplace=True, neg_slope=1e-2, norm_affine=True, conv_bias=False):
        super(ConvDropoutNormLReLU, self).__init__()
        self.conv = conv_op(input_channels, output_channels, kernel_size, stride, padding, dilation, bias=conv_bias)
        if dropout is not None and dropout != 0:
            self.dropout = dropout_op(dropout, inplace)
        else:
            self.dropout = None
        self.instnorm = norm_op(output_channels, affine=norm_affine)
        self.lrelu = nn.LeakyReLU(neg_slope, inplace=inplace)

    def forward(self, x):
        x = self.conv(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return self.lrelu(self.instnorm(x))


class StackedConvLayers(nn.Module):
    def __init__(self, input_feature_channels, output_feature_channels, num_convs, dilation=1, kernel_size=3, padding=1, dropout=0.0, initial_stride=1, conv_op=nn.Conv2d,
                 norm_op=nn.BatchNorm2d, dropout_op=nn.Dropout2d, inplace=True, neg_slope=1e-2, norm_affine=True, conv_bias=False):
        '''
        stacks ConvDropoutNormLReLU layers. initial_stride will only be applied to first layer in the stack. The other parameters affect all layers
        :param input_feature_channels:
        :param output_feature_channels:
        :param num_convs:
        :param dilation:
        :param kernel_size:
        :param padding:
        :param dropout:
        :param initial_stride:
        :param conv_op:
        :param norm_op:
        :param dropout_op:
        :param inplace:
        :param neg_slope:
        :param norm_affine:
        :param conv_bias:
        '''
        self.input_channels = input_feature_channels
        self.output_channels = output_feature_channels
        super(StackedConvLayers, self).__init__()
        self.blocks = nn.Sequential(*([ConvDropoutNormLReLU(input_feature_channels, output_feature_channels, dilation, kernel_size, padding, dropout, initial_stride, conv_op, norm_op, dropout_op, inplace, neg_slope, norm_affine, conv_bias)]
                                      + [ConvDropoutNormLReLU(output_feature_channels, output_feature_channels, dilation, kernel_size, padding, dropout, 1, conv_op, norm_op, dropout_op, inplace, neg_slope, norm_affine, conv_bias)] * (
                                          num_convs - 1)))

    def forward(self, x):
        return self.blocks(x)


class Generic_UNet(nn.Module):
    def __init__(self, input_channels, base_num_features, num_classes, dropout_p, num_conv_per_stage=2, num_downscale=4,
                 feat_map_mul_on_downscale=2, conv_op=nn.Conv2d, norm_op=nn.InstanceNorm2d, dropout_op=nn.Dropout2d,
                 conv_bias=False, inplace=True, neg_slope=1e-2, norm_affine=True, deep_supervision=True,
                 dropout_in_localization=False, final_nonlin=softmax_helper):
        super(Generic_UNet, self).__init__()
        self.final_nonlin = final_nonlin
        assert num_conv_per_stage > 1, "this implementation does not support only one conv per stage"

        if conv_op == nn.Conv2d:
            upsample_mode = 'bilinear'
            pool_op = nn.MaxPool2d
        elif conv_op == nn.Conv3d:
            upsample_mode = 'trilinear'
            pool_op = nn.MaxPool3d
        else:
            raise ValueError("unknown convolution dimensionality, conv op: %s" % str(conv_op))

        self.do_ds = deep_supervision
        self.conv_blocks_context = []
        self.conv_blocks_localization = []
        self.td = []
        self.tu = []
        self.seg_outputs = []

        output_features = base_num_features
        input_features = input_channels
        for d in range(num_downscale):
            self.conv_blocks_context.append(StackedConvLayers(input_features, output_features, num_conv_per_stage, 1, 3, 1, dropout_p, 1, conv_op, norm_op, dropout_op, inplace, neg_slope, norm_affine, conv_bias))
            self.td.append(pool_op(2))
            input_features = output_features
            output_features = int(np.round(output_features * feat_map_mul_on_downscale))

        self.conv_blocks_context.append(nn.Sequential(
            StackedConvLayers(input_features, output_features, num_conv_per_stage - 1, 1, 3, 1, dropout_p, 1, conv_op, norm_op, dropout_op, inplace, neg_slope, norm_affine, conv_bias),
            StackedConvLayers(output_features, self.conv_blocks_context[-1].output_channels, 1, 1, 3, 1, dropout_p, 1, conv_op, norm_op, dropout_op, inplace, neg_slope, norm_affine, conv_bias)))

        if not dropout_in_localization:
            dropout_p = 0.0

        for u in range(num_downscale):
            input_features = self.conv_blocks_context[-(2 + u)].output_channels * 2
            output_features = self.conv_blocks_context[-(2 + u)].output_channels
            if u != num_downscale-1:
                final_output_size = self.conv_blocks_context[-(3 + u)].output_channels
            else:
                final_output_size = output_features
            self.tu.append(nn.Upsample(scale_factor=2, mode=upsample_mode))
            self.conv_blocks_localization.append(nn.Sequential(
                StackedConvLayers(input_features, output_features, num_conv_per_stage - 1, 1, 3, 1, dropout_p, 1,
                                  conv_op, norm_op, dropout_op, inplace, neg_slope, norm_affine, conv_bias),
                StackedConvLayers(output_features, final_output_size, 1, 1, 3, 1, dropout_p,
                                  1, conv_op, norm_op, dropout_op, inplace, neg_slope, norm_affine, conv_bias)
            ))

        for ds in range(len(self.conv_blocks_localization)):
            self.seg_outputs.append(conv_op(self.conv_blocks_localization[ds][-1].output_channels, num_classes, 1, 1, 0, 1, 1, False))

        # register all modules properly
        self.conv_blocks_localization = nn.ModuleList(self.conv_blocks_localization)
        self.conv_blocks_context = nn.ModuleList(self.conv_blocks_context)
        self.td = nn.ModuleList(self.td)
        self.tu = nn.ModuleList(self.tu)
        self.seg_outputs = nn.ModuleList(self.seg_outputs)

        self.apply(init_weights)

    def forward(self, x):
        skips = []
        seg_outputs = []
        for d in range(len(self.conv_blocks_context) - 1):
            x = self.conv_blocks_context[d](x)
            skips.append(x)
            x = self.td[d](x)

        x = self.conv_blocks_context[-1](x)

        for u in range(len(self.tu)):
            x = self.tu[u](x)
            x = torch.cat((x, skips[-(u + 1)]), dim=1)
            x = self.conv_blocks_localization[u](x)
            seg_outputs.append(self.final_nonlin(self.seg_outputs[u](x)))

        if self.do_ds:
            return tuple(seg_outputs[::-1])
        else:
            return seg_outputs[-1]
